/*
 * Copyright (c) 2003 Regents of The University of Michigan.
 * All Rights Reserved.  See COPYRIGHT.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "argcargv.h"
#include "code.h"
#include "mkdirs.h"
#include "pathcmp.h"

int		cksum = 1;
int		verbose = 0;
int		noupload = 0;
extern char   	*version;

struct node {
    char                path[ MAXPATHLEN ];
    struct node         *next;
};

struct node* create_node( char *path );
void free_node( struct node *node );

   struct node *
create_node( char *path )
{
    struct node         *new_node;

    if (( new_node = (struct node *) malloc( sizeof( struct node ))) == NULL ) {
	perror( "malloc" );
	return( NULL );
    }
    if ( snprintf( new_node->path, MAXPATHLEN, "%s", path ) >= MAXPATHLEN ) {
	fprintf( stderr, "%s: path too long\n", path );
	return( NULL );
    }

    return( new_node );
}

    void
free_node( struct node *node )
{
    free( node );
}

struct tran {
    int                 num;
    int                 eof;
    int                 tac;
    int                 remove;
    int                 linenum;
    char                *path;
    char                *name;
    char                *line;
    char                tline[ 2 * MAXPATHLEN ];
    char                prepath[ MAXPATHLEN ];
    char                filepath[ MAXPATHLEN ];
    char                **targv;
    FILE                *fs;
    ACAV                *acav;
    struct node         *next;
};

int getnextline( struct tran *tran ); 

    int
getnextline( struct tran *tran )
{
    int len;

getline:
    if ( fgets( tran->tline, MAXPATHLEN, tran->fs ) == NULL ) {
	tran->eof = 1;
	return( 0 );
    }
    tran->linenum++;

    if ( tran->line != NULL ) {
	free( tran->line );
	tran->line = NULL;
    }

    if ( ( tran->line = strdup( tran->tline ) ) == NULL ) {
	perror( tran->tline );
	return( -1 );
    }

    /* Check line length */
    len = strlen( tran->tline );
    if ( ( tran->tline[ len - 1 ] ) != '\n' ) {
	fprintf( stderr, "%s: %d: %s: line too long\n", tran->name,
	    tran->linenum, tran->tline );
	return( -1 );
    }
    if ( ( tran->tac = acav_parse( tran->acav,
	    tran->tline, &(tran->targv) )  ) < 0 ) {
	fprintf( stderr, "acav_parse\n" );
	return( -1 );
    }
    /* Skip blank lines and comments */
    if (( tran->tac == 0 ) || ( *tran->targv[ 0 ] == '#' )) {
	goto getline;
    }

    if ( *tran->targv[ 0 ] == '-' ) {
	tran->remove = 1;
	tran->targv++;
    } else {
	tran->remove = 0;
    }

    /* Decode file path */
    if ( snprintf( tran->filepath, MAXPATHLEN, "%s", decode( tran->targv[ 1 ]))
	> MAXPATHLEN - 1 ) {
	fprintf( stderr, "%s: line %d: %s: path too long\n",
		tran->name, tran->linenum, decode( tran->targv[ 1 ]));
	return( 1 );
    }

    /* Check transcript order */
    if ( tran->prepath != 0 ) {
	if ( pathcmp( tran->filepath, tran->prepath ) < 0 ) {
	    fprintf( stderr, "%s: line %d: bad sort order\n",
			tran->name, tran->linenum );
	    return( 1 );
	}
    }
    if ( snprintf( tran->prepath, MAXPATHLEN, "%s", tran->filepath )
	    >= MAXPATHLEN ) { 
	fprintf( stderr, "%s: line %d: %s: path too long\n",
		tran->name, tran->linenum, tran->filepath );
	return( 1 );
    }

    return( 0 );
}

/*
 * exit codes:
 *	0  	okay	
 *	2	System error
 */

    int
main( int argc, char **argv )
{
    int			c, i, j, cmpval, err = 0, tcount = 0, candidate = NULL;
    int			force = 0, ofd, fileloc = 0, match = 0;
    char		*tname = NULL, *file = NULL;
    char		*tpath = NULL;
    char		npath[ 2 * MAXPATHLEN ];
    char		opath[ 2 * MAXPATHLEN ];
    struct tran		**trans = NULL;
    struct node		*new_node = NULL;
    struct node		*node = NULL;
    struct node		*dirlist = NULL;
    FILE		*ofs;
    mode_t		mask;

    while ( ( c = getopt( argc, argv, "fnu:Vv" ) ) != EOF ) {
	switch( c ) {
	case 'f':
	    force = 1;
	    break;
	case 'n':
	    noupload = 1;
	    break;
	case 'u':
	    errno = 0;
	    mask = (mode_t)strtol( optarg, (char **)NULL, 0 );
	    if ( errno != 0 ) {
		err++;
		break;
	    }
	    umask( mask );
	    break;
	case 'V':
	    printf( "%s\n", version );
	    exit( 0 );
	case 'v':
	    verbose = 1;
	    break;
	default:
	    err++;
	    break;
	}
    }

    tcount = argc - ( optind + 1 );	/* "+ 1" accounts for dest tran */

    if ( noupload && ( tcount > 2 ) ) {
	err++;
    }
    if ( force && ( tcount > 1 ) ) {
	err++;
    }
    if ( !force && ( tcount < 2 )) {
	err++;
    }

    if ( err ) {
	fprintf( stderr, "Usage: %s [-vV] [ -u umask ] ", argv[ 0 ] );
	fprintf( stderr, "transcript... dest\n" );
	fprintf( stderr, "       %s -f [-vV] [ -u umask ] ", argv[ 0 ] );
	fprintf( stderr, "transcript1 transcript2\n" );
	fprintf( stderr, "       %s -n [-vV] [ -u umask ] ", argv[ 0 ] );
	fprintf( stderr, "transcript1 transcript2 dest\n" );
	exit( 2 );
    }

    tpath = argv[ argc - 1 ];
    if ( force ) {
	/* Check for write access */
	if ( access( argv[ argc - 1 ], W_OK ) != 0 ) {
	    perror( argv[ argc - 1 ] );
	    exit( 2 );
	}
	tcount++;			/* add dest to tran merge list */
    }

    /* Create array of transcripts */
    if ( ( trans = (struct tran**)malloc(
	    sizeof( struct tran* ) * ( tcount ) ) ) == NULL ) {
	perror( "malloc" );
	exit( 2 );
    }

    /* loop over array of trans */
    for ( i = 0;  i < tcount;  i++ ) {

	if ( ( trans[ i ] = (struct tran*)malloc( sizeof( struct tran ) ) )
		== NULL ) {
	    perror( "malloc" );
	    return( 1 );
	}
	trans[ i ]->num = i;
	trans[ i ]->eof = 0;
	trans[ i ]->linenum = 0;
	*trans[ i ]->prepath = 0;

	/* open tran */
	if ( ( trans[ i ]->fs = fopen( argv[ i + optind ], "r" ) ) == NULL ) {
	    perror( argv[ i + optind ] );
	    return( 1 );
	}

	/* Get transcript name from path */
	trans[ i ]->path = argv[ i + optind ];
	if ( ( trans[ i ]->name = strrchr( trans[ i ]->path, '/' ) ) == NULL ) {
	    trans[ i ]->name = trans[ i ]->path;
	    trans[ i ]->path = ".";
	} else {
	    *trans[ i ]->name = (char)'\0';
	    trans[ i ]->name++;
	}

	if ( ( trans[ i ]->acav = acav_alloc() ) == NULL ) {
	    fprintf( stderr, "acav_malloc\n" );
	    return( 1 );
	}
	trans[ i ]->line = NULL;
	if ( getnextline( trans[ i ] ) < 0 ) {
	    exit( 2 );
	}
    }

    if ( force ) {
	tname = trans[ 1 ]->name;
	tpath = trans[ 1 ]->path;
    } else {
	/* Get new transcript name from transcript path */
	if ( ( tname = strrchr( tpath, '/' ) ) == NULL ) {
	    tname = tpath;
	    tpath = ".";
	} else {
	    *tname = (char)'\0';
	    tname++;
	}
    }

    if ( !force ) {
	/* Create file/tname dir */
	if ( snprintf( npath, MAXPATHLEN, "%s/../file/%s.%d", tpath, tname,
		(int)getpid()) > MAXPATHLEN -1 ) {
	    fprintf( stderr, "%s/../file/%s.%d: path too long\n", tpath,
		tname, (int)getpid());
	    exit( 2 );
	}
	if ( mkdir( npath, (mode_t)0777 ) != 0 ) {
	    perror( npath );
	    exit( 2 );
	}
    }

    /* Create temp transcript/tname file */
    if ( snprintf( opath, MAXPATHLEN, "%s/%s.%d", tpath, tname, (int)getpid())
	    > MAXPATHLEN - 1 ) {
	fprintf( stderr, "%s/%s.%d: path too long\n", tpath, tname,
	    (int)getpid());
	exit( 2 );
    }
    if ( ( ofd = open( opath, O_WRONLY | O_CREAT | O_EXCL,
	    0666 ) ) < 0 ) {
	perror( opath );
	exit( 2 );
    }
    if ( ( ofs = fdopen( ofd, "w" ) ) == NULL ) {
	perror( opath );
	exit( 2 );
    }
    
    /* merge */
    for ( i = 0; i < tcount; i++ ) {
	while ( !(trans[ i ]->eof)) {
	    candidate = i;
	    fileloc = i;

	    if ( force && ( candidate == ( tcount - 1 ))) {
		goto outputline;
	    }

	    /* Compare candidate to other transcripts */
	    match = 0;
	    for ( j = i + 1; j < tcount; j++ ) {
		if ( trans[ j ]->eof ) {
		    continue;
		}
		cmpval = pathcmp( trans[ candidate ]->filepath,
		    trans[ j ]->filepath );
		if ( cmpval == 0 ) {
		    /* File match */
		    match = 1;

		    if (( noupload ) &&
			    ( *trans[ candidate ]->targv[ 0 ] == 'f' 
			    || *trans[ candidate ]->targv[ 0 ] == 'a' )) {
			/* Use lower precedence path */
			trans[ candidate ]->path = 
			    trans[ j ]->path;

			/* Select which file should be linked */
			if ( ( strcmp( trans[ candidate ]->targv[ 6 ], 
				trans[ j ]->targv[ 6 ] ) == 0 ) &&
				( strcmp( trans[ candidate ]->targv[ 7 ],
				trans[ j ]->targv[ 7 ] ) == 0 ) ) {
			    fileloc = j;
			} else {
			    /* don't print file only in highest tran */
			    goto skipline;
			}
		    }
		    if ( ( force ) && ( *trans[ j ]->targv[ 0 ] == 'f' 
			    || *trans[ j ]->targv[ 0 ] == 'a' )) {
			/* Remove file from lower precedence transcript */
			if ( snprintf( opath, MAXPATHLEN, "%s/../file/%s/%s",
				trans[ j ]->path, trans[ j ]->name,
				trans[ j ]->filepath ) > MAXPATHLEN -1 ) {
			    fprintf( stderr,
				"%s/../file/%s/%s: path too long\n",
				trans[ j ]->path, trans[ j ]->name,
				trans[ j ]->filepath );
			    exit( 2 );
			}
			if ( unlink( opath ) != 0 ) {
			    perror( opath );
			    exit( 2 );
			}
			if ( verbose ) printf( "%s: %s: unlinked\n",
			    trans[ j ]->name, trans[ j ]->filepath);
		    }
		    /* Advance lower precedence transcript */
		    if ( getnextline( trans[ j ] ) < 0 ) {
			exit( 2 );
		    }
		} else if ( cmpval > 0 ) {
		    candidate = j;
		    fileloc = j;
		}
	    }
	    if ( force && ( candidate == 1 ) ) {
		goto outputline;
	    }
	    /* skip items to be removed or files not uploaded */
	    if (( trans[ candidate ]->remove ) ||
		    (( noupload ) && ( candidate == 0 ) && ( fileloc == 0 ))) {
		if ( match && force &&
			( *trans[ candidate ]->targv[ 0 ] == 'd' )) {
		    new_node = create_node( trans[ candidate ]->targv[ 1 ] );
		    new_node->next = dirlist;
		    dirlist = new_node;
		}
		goto skipline;
	    }
	    /* output non-files */
	    if ( *trans[ candidate ]->targv[ 0 ] != 'f'
		    && *trans[ candidate ]->targv[ 0 ] != 'a' ) {
		goto outputline;
	    }

	    /*
	     * Assume that directory structure is present so the entire path
	     * is not recreated for every file.  Only if link fails is
	     * mkdirs() called.
	     */
	    /* Fix for broken mkdir on OS X */
	    if ( *trans[ candidate ]->filepath == '/' ) {
		if ( snprintf( opath, MAXPATHLEN, "%s/../file/%s%s",
			trans[ candidate ]->path, trans[ fileloc ]->name,
			trans[ candidate ]->filepath ) > MAXPATHLEN - 1 ) {
		    fprintf( stderr, "%s/../file/%s/%s: path too long\n",
			trans[ candidate ]->path, trans[ fileloc ]->name,
			trans[ candidate ]->filepath );
		    exit( 2 );
		}
	    } else {
		if ( snprintf( opath, MAXPATHLEN, "%s/../file/%s/%s",
			trans[ candidate ]->path, trans[ fileloc ]->name,
			trans[ candidate ]->filepath ) > MAXPATHLEN - 1 ) {
		    fprintf( stderr, "%s/../file/%s/%s: path too long\n",
			trans[ candidate ]->path, trans[ fileloc ]->name,
			trans[ candidate ]->filepath );
		    exit( 2 );
		}
	    }

	    if ( !force ) {
		/* Fix for broken mkdir on OS X */
		if ( *trans[ candidate ]->filepath == '/' ) {
		    if ( snprintf( npath, MAXPATHLEN, "%s/../file/%s.%d/%s",
			    tpath, tname, (int)getpid(),
			    trans[ candidate ]->filepath ) > MAXPATHLEN - 1 ) {
			fprintf( stderr, "%s/../file/%s.%d/%s: path too long\n",
			    tpath, tname, (int)getpid(),
			    trans[ candidate ]->filepath );
			exit( 2 );
		    }
		} else {
		    if ( snprintf( npath, MAXPATHLEN, "%s/../file/%s.%d/%s",
			    tpath, tname, (int)getpid(),
			    trans[ candidate ]->filepath ) > MAXPATHLEN - 1 ) {
			fprintf( stderr, "%s/../file/%s.%d/%s: path too long\n",
			    tpath, tname, (int)getpid(),
			    trans[ candidate ]->filepath );
			exit( 2 );
		    }
		}
	    } else {
		if ( snprintf( npath, MAXPATHLEN, "%s/../file/%s/%s", tpath,
			tname, trans[ candidate ]->filepath )
			> MAXPATHLEN - 1 ) {
		    fprintf( stderr, "%s/../file/%s/%s: path too long\n", 
			tpath, tname, trans[ candidate ]->filepath );
		    exit( 2 );
		}
	    }

	    /* First try to link file */
	    if ( link( opath, npath ) != 0 ) {

		/* If that fails, verify directory structure */
		if ( ( file = strrchr( trans[ candidate ]->targv[ 1 ], '/' ) )
			!= NULL ) {
		    if ( !force ) {
			/* Fix for broken mkdir on OS X */
			if ( *trans[ candidate ]->filepath == '/' ) {
			    if ( snprintf( npath, MAXPATHLEN,
				    "%s/../file/%s.%d%s",
				    tpath, tname, (int)getpid(), 
				    trans[ candidate ]->filepath )
				    > MAXPATHLEN - 1 ) {
				fprintf( stderr,
				    "%s/../file/%s.%d%s: path too long\n",
				    tpath, tname, (int)getpid(),
				    trans[ candidate ]->filepath );
				exit( 2 );
			    }
			} else {
			    if ( snprintf( npath, MAXPATHLEN,
				    "%s/../file/%s.%d/%s",
				    tpath, tname, (int)getpid(), 
				    trans[ candidate ]->filepath )
				    > MAXPATHLEN - 1 ) {
				fprintf( stderr,
				    "%s/../file/%s.%d/%s: path too long\n",
				    tpath, tname, (int)getpid(),
				    trans[ candidate ]->filepath );
				exit( 2 );
			    }
			}
		    } else {
			if ( *trans[ candidate ]->filepath == '/' ) {
			    if ( snprintf( npath, MAXPATHLEN,
				    "%s/../file/%s%s",
				    tpath, tname, trans[ candidate ]->filepath )
				    > MAXPATHLEN - 1 ) {
				fprintf( stderr,
				    "%s/../file/%s%s: path too long\n",
				    tpath, tname,
				    trans[ candidate ]->filepath );
				exit( 2 );
			    }
			} else {
			    if ( snprintf( npath, MAXPATHLEN,
				    "%s/../file/%s/%s",
				    tpath, tname, trans[ candidate ]->filepath )
				    > MAXPATHLEN - 1 ) {
				fprintf( stderr,
				    "%s/../file/%s/%s: path too long\n",
				    tpath, tname,
				    trans[ candidate ]->filepath );
				exit( 2 );
			    }
			}
		    }
		    if ( mkdirs( npath ) != 0 ) {
			fprintf( stderr, "%s: mkdirs failed\n", npath );
			exit( 2 );
		    }
		} 

		/* Try link again */
		if ( link( opath, npath ) != 0 ) {
		    fprintf( stderr, "linking %s -> %s: %s",
			opath, npath, strerror( errno ));
		    exit( 2 );
		}
	    }
	    if ( verbose ) printf( "%s: %s: merged into: %s\n",
		trans[ candidate ]->name, trans[ candidate ]->filepath,
		tname );
		
outputline:
	    /* Output line */
	    if ( fputs( trans[ candidate ]->line, ofs ) == EOF ) {
		perror( trans[ candidate ]->line );
		exit( 2 );
	    }
skipline:
	    if (( trans[ candidate ]->remove ) && !match ) {
		/* Recreate unmatched "-" line */
		if ( fputs( trans[ candidate ]->line, ofs ) == EOF ) {
		    perror( trans[ candidate ]->line );
		    exit( 2 );

		}
	    }
	    if ( getnextline( trans[ candidate ] ) != 0 ) {
		exit( 2 );
	    }
	}
    }

    if ( force ) {
	while ( dirlist != NULL ) {
	    node = dirlist;
	    dirlist = node->next;
	    if ( snprintf( opath, MAXPATHLEN, "%s/../file/%s/%s", tpath,
		    tname, node->path ) >= MAXPATHLEN ) {
		fprintf( stderr, "%s/../file/%s/%s: path too long\n", 
		    tpath, tname, node->path );
		exit( 2 );
	    }
	    if ( rmdir( opath ) != 0 ) {
		if (( errno == EEXIST ) || ( errno == ENOTEMPTY )) {
		    fprintf( stderr, "%s: %s: Not empty, continuing...\n",
			tname, node->path );
		} else if ( errno != ENOENT ) {
		    perror( opath );
		    exit( 2 );
		}
	    } else {
		if ( verbose ) printf( "%s: %s: unlinked\n", tname,
		    node->path );
	    }
	    free_node( node );
	}
    }

    /* Rename temp transcript and file structure */
    if ( !force ) {
	if ( snprintf( opath, MAXPATHLEN, "%s/../file/%s.%d", tpath,
		tname, (int)getpid()) > MAXPATHLEN - 1 ) {
	    fprintf( stderr, "%s/../file/%s.%d: path too long\n",
		tpath, tname, (int)getpid());
	    exit( 2 );
	}
	if ( snprintf( npath, MAXPATHLEN, "%s/../file/%s", tpath, tname )
		> MAXPATHLEN - 1 ) {
	    fprintf( stderr, "%s/../file/%s: path too long\n", tpath, tname );
	    exit( 2 );
	}
	if ( rename( opath, npath ) != 0 ) {
	    perror( npath );
	    exit( 2 );
	}
    }
    if ( snprintf( opath, MAXPATHLEN, "%s/%s.%d", tpath, tname, (int)getpid())
	    > MAXPATHLEN - 1 ) {
	fprintf( stderr, "%s/%s.%d: path too long\n", tpath, tname,
	    (int)getpid());
	exit( 2 );
    }
    if ( snprintf( npath, MAXPATHLEN, "%s/%s", tpath, tname )
	    > MAXPATHLEN - 1 ) {
	fprintf( stderr, "%s/%s: path too long\n", tpath, tname );
	exit( 2 );
    }

    if ( rename( opath, npath ) != 0 ) {
	perror( npath );
	exit( 2 );
    }

    exit( 0 );
} 
